=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2006-09-21 09:15:20 +0000
+++ bzrlib/errors.py	2006-09-22 04:39:58 +0000
@@ -266,6 +266,14 @@
     """Directory not empty: %(path)r%(extra)s"""
 
 
+class ReadingCompleted(BzrNewError):
+    """The MediumRequest '%(request)s' has already had finish_reading called upon it - the request has been completed and no more data may be read."""
+
+    def __init__(self, request):
+        BzrNewError.__init__(self)
+        self.request = request
+
+
 class ResourceBusy(PathError):
     """Device or resource busy: %(path)r%(extra)s"""
 
@@ -791,11 +799,21 @@
 
 class NoSuchExportFormat(BzrNewError):
     """Export format %(format)r not supported"""
+
     def __init__(self, format):
         BzrNewError.__init__(self)
         self.format = format
 
 
+
+class TooManyConcurrentRequests(BzrNewError):
+    """The medium '%(medium)s' has reached its concurrent request limit. Be sure to finish_writing and finish_reading on the current request that is open."""
+
+    def __init__(self, medium):
+        BzrNewError.__init__(self)
+        self.medium = medium
+
+
 class TransportError(BzrNewError):
     """Transport error: %(msg)s %(orig_error)s"""
 
@@ -892,6 +910,22 @@
                           " unchanged." % tree.basedir)
 
 
+class WritingCompleted(BzrNewError):
+    """The MediumRequest '%(request)s' has already had finish_writing called upon it - accept bytes may not be called anymore."""
+
+    def __init__(self, request):
+        BzrNewError.__init__(self)
+        self.request = request
+
+
+class WritingNotComplete(BzrNewError):
+    """The MediumRequest '%(request)s' has not has finish_writing called upon it - until the write phase is complete no data may be read."""
+
+    def __init__(self, request):
+        BzrNewError.__init__(self)
+        self.request = request
+
+
 class CantReprocessAndShowBase(BzrNewError):
     """Can't reprocess and show base.
 Reprocessing obscures relationship of conflicting lines to base."""

=== modified file 'bzrlib/tests/test_errors.py'
--- bzrlib/tests/test_errors.py	2006-09-21 09:15:20 +0000
+++ bzrlib/tests/test_errors.py	2006-09-22 04:39:58 +0000
@@ -49,6 +49,13 @@
             "atree.",
             str(error))
 
+    def test_too_many_concurrent_requests(self):
+        error = errors.TooManyConcurrentRequests("a medium")
+        self.assertEqualDiff("The medium 'a medium' has reached its concurrent "
+            "request limit. Be sure to finish_writing and finish_reading on "
+            "the current request that is open.",
+            str(error))
+
     def test_up_to_date(self):
         error = errors.UpToDateFormat(bzrdir.BzrDirFormat4())
         self.assertEqualDiff("The branch format Bazaar-NG branch, "
@@ -64,6 +71,27 @@
                              repo.bzrdir.root_transport.base,
                              str(error))
 
+    def test_reading_completed(self):
+        error = errors.ReadingCompleted("a request")
+        self.assertEqualDiff("The MediumRequest 'a request' has already had "
+            "finish_reading called upon it - the request has been completed and"
+            " no more data may be read.",
+            str(error))
+
+    def test_writing_completed(self):
+        error = errors.WritingCompleted("a request")
+        self.assertEqualDiff("The MediumRequest 'a request' has already had "
+            "finish_writing called upon it - accept bytes may not be called "
+            "anymore.",
+            str(error))
+
+    def test_writing_not_completed(self):
+        error = errors.WritingNotComplete("a request")
+        self.assertEqualDiff("The MediumRequest 'a request' has not has "
+            "finish_writing called upon it - until the write phase is complete"
+            " no data may be read.",
+            str(error))
+
 
 class PassThroughError(errors.BzrNewError):
     """Pass through %(foo)s and %(bar)s"""

=== modified file 'bzrlib/tests/test_smart_transport.py'
--- bzrlib/tests/test_smart_transport.py	2006-09-21 09:15:20 +0000
+++ bzrlib/tests/test_smart_transport.py	2006-09-22 04:39:58 +0000
@@ -102,9 +102,11 @@
         return t
     
     def test_construct_smart_stream_medium_client(self):
-        # make a new client; this really wants a connector function that returns
-        # two fifos or sockets but the constructor should not do any IO
-        medium = smart.SmartStreamMediumClient()
+        # make a new instance of the common base for Stream-like Mediums.
+        # this just ensures that the constructor stays parameter-free which
+        # is important for reuse : some subclasses will dynamically connect,
+        # others are always on, etc.
+        medium = smart.SmartClientStreamMedium()
 
     def test_construct_smart_client_medium(self):
         # the base client medium takes no parameters
@@ -116,12 +118,34 @@
         # Constructing one should just save these and do nothing.
         # We test this by passing in None.
         medium = smart.SmartSimplePipesClientMedium(None, None)
-
-    def test_simple_pipes_client_accept_bytes_writes_to_writable(self):
+        
+    def test_simple_pipes_client_request_type(self):
+        # SimplePipesClient should use SmartClientStreamMediumRequest's.
+        medium = smart.SmartSimplePipesClientMedium(None, None)
+        request = medium.get_request()
+        self.assertIsInstance(request, smart.SmartClientStreamMediumRequest)
+
+    def test_simple_pipes_client_get_concurrent_requests(self):
+        # the simple_pipes client does not support pipelined requests:
+        # but it does support serial requests: we construct one after 
+        # another is finished. This is a smoke test testing the integration
+        # of the SmartClientStreamMediumRequest and the SmartClientStreamMedium
+        # classes - as the sibling classes share this logic, they do not have
+        # explicit tests for this.
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(None, output)
+        request = medium.get_request()
+        request.finished_writing()
+        request.finished_reading()
+        request2 = medium.get_request()
+        request2.finished_writing()
+        request2.finished_reading()
+
+    def test_simple_pipes_client__accept_bytes_writes_to_writable(self):
         # accept_bytes writes to the writeable pipe.
         output = StringIO()
         medium = smart.SmartSimplePipesClientMedium(None, output)
-        medium.accept_bytes('abc')
+        medium._accept_bytes('abc')
         self.assertEqual('abc', output.getvalue())
     
     def test_simple_pipes_client_disconnect_does_nothing(self):
@@ -131,7 +155,7 @@
         medium = smart.SmartSimplePipesClientMedium(input, output)
         # send some bytes to ensure disconnecting after activity still does not
         # close.
-        medium.accept_bytes('abc')
+        medium._accept_bytes('abc')
         medium.disconnect()
         self.assertFalse(input.closed)
         self.assertFalse(output.closed)
@@ -142,9 +166,9 @@
         input = StringIO()
         output = StringIO()
         medium = smart.SmartSimplePipesClientMedium(input, output)
-        medium.accept_bytes('abc')
+        medium._accept_bytes('abc')
         medium.disconnect()
-        medium.accept_bytes('abc')
+        medium._accept_bytes('abc')
         self.assertFalse(input.closed)
         self.assertFalse(output.closed)
         self.assertEqual('abcabc', output.getvalue())
@@ -163,6 +187,24 @@
         self.assertEqual('abc', medium.read_bytes(3))
         medium.disconnect()
         self.assertEqual('def', medium.read_bytes(3))
+        
+    def test_simple_pipes_client_supports__flush(self):
+        # invoking _flush on a SimplePipesClient should flush the output 
+        # pipe. We test this by creating an output pipe that records
+        # flush calls made to it.
+        from StringIO import StringIO # get regular StringIO
+        input = StringIO()
+        output = StringIO()
+        flush_calls = []
+        def _flush(): flush_calls.append('flush')
+        output.flush = _flush
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        # this call is here to ensure we only flush once, not on every
+        # _accept_bytes call.
+        medium._accept_bytes('abc')
+        medium._flush()
+        medium.disconnect()
+        self.assertEqual(['flush'], flush_calls)
 
     def test_construct_smart_ssh_client_medium(self):
         # the SSH client medium takes:
@@ -186,7 +228,7 @@
         vendor = StringIOSSHVendor(StringIO(), output)
         medium = smart.SmartSSHClientMedium('a hostname', 'a port', 'a username',
             'a password', vendor)
-        medium.accept_bytes('abc')
+        medium._accept_bytes('abc')
         self.assertEqual('abc', output.getvalue())
         self.assertEqual([('connect_ssh', 'a username', 'a password',
             'a hostname', 'a port',
@@ -205,7 +247,7 @@
         os.environ['BZR_REMOTE_PATH'] = 'fugly'
         medium = smart.SmartSSHClientMedium('a hostname', 'a port', 'a username',
             'a password', vendor)
-        medium.accept_bytes('abc')
+        medium._accept_bytes('abc')
         self.assertEqual('abc', output.getvalue())
         self.assertEqual([('connect_ssh', 'a username', 'a password',
             'a hostname', 'a port',
@@ -219,7 +261,7 @@
         output = StringIO()
         vendor = StringIOSSHVendor(input, output)
         medium = smart.SmartSSHClientMedium('a hostname', vendor=vendor)
-        medium.accept_bytes('abc')
+        medium._accept_bytes('abc')
         medium.disconnect()
         self.assertTrue(input.closed)
         self.assertTrue(output.closed)
@@ -239,7 +281,7 @@
         output = StringIO()
         vendor = StringIOSSHVendor(input, output)
         medium = smart.SmartSSHClientMedium('a hostname', vendor=vendor)
-        medium.accept_bytes('abc')
+        medium._accept_bytes('abc')
         medium.disconnect()
         # the disconnect has closed output, so we need a new output for the
         # new connection to write to.
@@ -247,7 +289,7 @@
         output2 = StringIO()
         vendor.read_from = input2
         vendor.write_to = output2
-        medium.accept_bytes('abc')
+        medium._accept_bytes('abc')
         medium.disconnect()
         self.assertTrue(input.closed)
         self.assertTrue(output.closed)
@@ -271,11 +313,30 @@
 
     def test_ssh_client_raises_on_read_when_not_connected(self):
         # Doing a read on a new (and thus unconnected) SSH medium raises
-        # MediumNotConnectedError.
+        # MediumNotConnected.
         medium = smart.SmartSSHClientMedium(None)
         self.assertRaises(errors.MediumNotConnected, medium.read_bytes, 0)
         self.assertRaises(errors.MediumNotConnected, medium.read_bytes, 1)
 
+    def test_ssh_client_supports__flush(self):
+        # invoking _flush on a SSHClientMedium should flush the output 
+        # pipe. We test this by creating an output pipe that records
+        # flush calls made to it.
+        from StringIO import StringIO # get regular StringIO
+        input = StringIO()
+        output = StringIO()
+        flush_calls = []
+        def _flush(): flush_calls.append('flush')
+        output.flush = _flush
+        vendor = StringIOSSHVendor(input, output)
+        medium = smart.SmartSSHClientMedium('a hostname', vendor=vendor)
+        # this call is here to ensure we only flush once, not on every
+        # _accept_bytes call.
+        medium._accept_bytes('abc')
+        medium._flush()
+        medium.disconnect()
+        self.assertEqual(['flush'], flush_calls)
+        
     def test_construct_smart_tcp_client_medium(self):
         # the TCP client medium takes a host and a port.  Constructing it won't
         # connect to anything.
@@ -291,7 +352,7 @@
         sock, medium = self.make_loopsocket_and_medium()
         bytes = []
         t = self.receive_bytes_on_server(sock, bytes)
-        medium.accept_bytes('abc')
+        medium._accept_bytes('abc')
         t.join()
         sock.close()
         self.assertEqual(['abc'], bytes)
@@ -303,7 +364,7 @@
         sock, medium = self.make_loopsocket_and_medium()
         bytes = []
         t = self.receive_bytes_on_server(sock, bytes)
-        medium.accept_bytes('ab')
+        medium._accept_bytes('ab')
         medium.disconnect()
         t.join()
         sock.close()
@@ -320,19 +381,134 @@
 
     def test_tcp_client_raises_on_read_when_not_connected(self):
         # Doing a read on a new (and thus unconnected) TCP medium raises
-        # MediumNotConnectedError.
+        # MediumNotConnected.
         medium = smart.SmartTCPClientMedium(None, None)
         self.assertRaises(errors.MediumNotConnected, medium.read_bytes, 0)
         self.assertRaises(errors.MediumNotConnected, medium.read_bytes, 1)
-        
-    def test_get_request_from_medium(self):
-        # get a request and then call "I'm finished reading" method, then call
-        # "I'm finished writing" one.
-        medium = smart.SmartSimplePipesClientMedium(None, None)
-        request = medium.get_request()
-        request.finished_writing()
-        request.finished_reading()
-        
+
+    def test_tcp_client_supports__flush(self):
+        # invoking _flush on a TCPClientMedium should do something useful.
+        # RBC 20060922 not sure how to test/tell in this case.
+        sock, medium = self.make_loopsocket_and_medium()
+        bytes = []
+        t = self.receive_bytes_on_server(sock, bytes)
+        # try with nothing buffered
+        medium._flush()
+        medium._accept_bytes('ab')
+        # and with something sent.
+        medium._flush()
+        medium.disconnect()
+        t.join()
+        sock.close()
+        self.assertEqual(['ab'], bytes)
+        # now disconnect again : this should not do anything, if disconnection
+        # really did disconnect.
+        medium.disconnect()
+
+
+class TestSmartClientStreamMediumRequest(tests.TestCase):
+    """Tests the for SmartClientStreamMediumRequest.
+    
+    SmartClientStreamMediumRequest is a helper for the three stream based 
+    mediums: TCP, SSH, SimplePipes, so we only test it once, and then test that
+    those three mediums implement the interface it expects.
+    """
+
+    def test_accept_bytes_after_finished_writing_errors(self):
+        # calling accept_bytes after calling finished_writing raises 
+        # WritingCompleted to prevent bad assumptions on stream environments
+        # breaking the needs of message-based environments.
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(None, output)
+        request = smart.SmartClientStreamMediumRequest(medium)
+        request.finished_writing()
+        self.assertRaises(errors.WritingCompleted, request.accept_bytes, None)
+
+    def test_accept_bytes(self):
+        # accept bytes should invoke _accept_bytes on the stream medium.
+        # we test this by using the SimplePipes medium - the most trivial one
+        # and checking that the pipes get the data.
+        input = StringIO()
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        request = smart.SmartClientStreamMediumRequest(medium)
+        request.accept_bytes('123')
+        request.finished_writing()
+        request.finished_reading()
+        self.assertEqual('', input.getvalue())
+        self.assertEqual('123', output.getvalue())
+
+    def test_construct_sets_stream_request(self):
+        # constructing a SmartClientStreamMediumRequest on a StreamMedium sets
+        # the current request to the new SmartClientStreamMediumRequest
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(None, output)
+        request = smart.SmartClientStreamMediumRequest(medium)
+        self.assertTrue(medium._current_request is request)
+
+    def test_construct_while_another_request_active_throws(self):
+        # constructing a SmartClientStreamMediumRequest on a StreamMedium with
+        # a non-None _current_request raises TooManyConcurrentRequests.
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(None, output)
+        medium._current_request = "a"
+        self.assertRaises(errors.TooManyConcurrentRequests,
+            smart.SmartClientStreamMediumRequest, medium)
+
+    def test_finished_read_clears_current_request(self):
+        # calling finished_reading clears the current request from the requests
+        # medium
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(None, output)
+        request = smart.SmartClientStreamMediumRequest(medium)
+        request.finished_writing()
+        request.finished_reading()
+        self.assertEqual(None, medium._current_request)
+
+    def test_finished_read_before_finished_write_errors(self):
+        # calling finished_reading before calling finished_writing triggers a
+        # WritingNotComplete error.
+        medium = smart.SmartSimplePipesClientMedium(None, None)
+        request = smart.SmartClientStreamMediumRequest(medium)
+        self.assertRaises(errors.WritingNotComplete, request.finished_reading)
+        
+    def test_read_bytes(self):
+        # read bytes should invoke _read_bytes on the stream medium.
+        # we test this by using the SimplePipes medium - the most trivial one
+        # and checking that the data is supplied. Its possible that a 
+        # faulty implementation could poke at the pipe variables them selves,
+        # but we trust that this will be caught as it will break the integration
+        # smoke tests.
+        input = StringIO('321')
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        request = smart.SmartClientStreamMediumRequest(medium)
+        request.finished_writing()
+        self.assertEqual('321', request.read_bytes(3))
+        request.finished_reading()
+        self.assertEqual('', input.read())
+        self.assertEqual('', output.getvalue())
+
+    def test_read_bytes_before_finished_write_errors(self):
+        # calling read_bytes before calling finished_writing triggers a
+        # WritingNotComplete error because the Smart protocol is designed to be
+        # compatible with strict message based protocols like HTTP where the
+        # request cannot be submitted until the writing has completed.
+        medium = smart.SmartSimplePipesClientMedium(None, None)
+        request = smart.SmartClientStreamMediumRequest(medium)
+        self.assertRaises(errors.WritingNotComplete, request.read_bytes, None)
+
+    def test_read_bytes_after_finished_reading_errors(self):
+        # calling read_bytes after calling finished_reading raises 
+        # ReadingCompleted to prevent bad assumptions on stream environments
+        # breaking the needs of message-based environments.
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(None, output)
+        request = smart.SmartClientStreamMediumRequest(medium)
+        request.finished_writing()
+        request.finished_reading()
+        self.assertRaises(errors.ReadingCompleted, request.read_bytes, None)
+
 
 class RemoteTransportTests(tests.TestCaseWithTransport):
 
@@ -351,11 +527,6 @@
         t = self.get_transport()
         self.assertIsInstance(t, smart.SmartTransport)
 
-    def test_get_client_from_transport(self):
-        t = self.get_transport()
-        client = t.get_smart_client()
-        self.assertIsInstance(client, smart.SmartStreamClient)
-
     def test_get_medium_from_transport(self):
         """Remote transport has a medium always, which it can return."""
         t = self.get_transport()
@@ -477,7 +648,7 @@
         # we create a real connection not a loopback one, but it will use the
         # same server and pipes
         conn2 = self.transport.clone('.')
-        self.assertTrue(self.transport._client is conn2._client)
+        self.assertTrue(self.transport._medium is conn2._medium)
 
     def test__remote_path(self):
         self.assertEquals('/foo/bar',
@@ -618,59 +789,37 @@
         self.assertEqual('example.com', t._host)
 
 
-class FakeClient(smart.SmartStreamClient):
-    """Emulate a client for testing a transport's use of the client."""
-
-    def __init__(self):
-        smart.SmartStreamClient.__init__(self, None)
-        self._calls = []
-
-    def _call(self, *args):
-        self._calls.append(('_call', args))
-        return ('ok', )
-
-    def _recv_bulk(self):
-        return 'bar'
-
-
 class TestSmartTransport(tests.TestCase):
         
     def test_use_connection_factory(self):
         # We want to be able to pass a client as a parameter to SmartTransport.
-        client = FakeClient()
-        transport = smart.SmartTransport('bzr://localhost/', medium=client)
+        input = StringIO("ok\n3\nbardone\n")
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        transport = smart.SmartTransport('bzr://localhost/', medium=medium)
 
         # We want to make sure the client is used when the first remote
-        # method is called.  No method should have been called yet.
-        self.assertEqual([], client._calls)
+        # method is called.  No data should have been sent, or read.
+        self.assertEqual(0, input.tell())
+        self.assertEqual('', output.getvalue())
 
-        # Now call a method that should result in a single request.
+        # Now call a method that should result in a single request : as the
+        # transport makes its own protocol instances, we check on the wire.
+        # XXX: TODO: give the transport a protocol factory, which can make
+        # an instrumented protocol for us.
         self.assertEqual('bar', transport.get_bytes('foo'))
-        # The only call to _call should have been to get /foo.
-        self.assertEqual([('_call', ('get', '/foo'))], client._calls)
+        # only the needed data should have been sent/received.
+        self.assertEqual(13, input.tell())
+        self.assertEqual('get\x01/foo\n', output.getvalue())
 
     def test__translate_error_readonly(self):
         """Sending a ReadOnlyError to _translate_error raises TransportNotPossible."""
-        client = FakeClient()
-        transport = smart.SmartTransport('bzr://localhost/', medium=client)
+        medium = smart.SmartClientMedium()
+        transport = smart.SmartTransport('bzr://localhost/', medium=medium)
         self.assertRaises(errors.TransportNotPossible,
             transport._translate_error, ("ReadOnlyError", ))
 
 
-class InstrumentedClient(smart.SmartStreamClient):
-    """A smart client whose writes are stored to a supplied list."""
-
-    def __init__(self, write_output_list):
-        smart.SmartStreamClient.__init__(self, None)
-        self._write_output_list = write_output_list
-
-    def _ensure_connection(self):
-        """We are never strictly connected."""
-
-    def _write_and_flush(self, bytes):
-        self._write_output_list.append(bytes)
-
-
 class InstrumentedServerProtocol(smart.SmartServerStreamMedium):
     """A smart server which is backed by memory and saves its write requests."""
 
@@ -698,10 +847,14 @@
 
     def setUp(self):
         super(TestSmartProtocol, self).setUp()
-        self.to_server = []
-        self.to_client = []
-        self.smart_client = InstrumentedClient(self.to_server)
-        self.smart_server = InstrumentedServerProtocol(self.to_client)
+        self.server_to_client = []
+        self.to_server = StringIO()
+        self.to_client = StringIO()
+        self.client_medium = smart.SmartSimplePipesClientMedium(self.to_client,
+            self.to_server)
+        self.client_protocol = smart.SmartClientRequestProtocolOne(
+            self.client_medium)
+        self.smart_server = InstrumentedServerProtocol(self.server_to_client)
         self.smart_server_request = smart.SmartServerRequestHandler(None)
 
     def assertOffsetSerialisation(self, expected_offsets, expected_serialised,
@@ -723,10 +876,20 @@
         serialised = client._serialise_offsets(offsets)
         self.assertEqual(expected_serialised, serialised)
 
-    def test_protocol_construction(self):
+    def test_construct_version_one_server_protocol(self):
         protocol = smart.SmartServerRequestProtocolOne(None, None)
-        self.assertFalse(protocol.finished_reading)
-        self.assertEqual(None, protocol.request)
+        self.assertEqual('', protocol.in_buffer)
+        self.assertFalse(protocol.has_dispatched)
+        # Once refactoring is complete, we don't need these assertions
+        self.assertFalse(hasattr(protocol, '_in'))
+        self.assertFalse(hasattr(protocol, '_out'))
+
+    def test_construct_version_one_client_protocol(self):
+        # we can construct a client protocol from a client medium request
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(None, output)
+        request = medium.get_request()
+        client_protocol = smart.SmartClientRequestProtocolOne(request)
 
     def test_server_offset_serialisation(self):
         """The Smart protocol serialises offsets as a comma and \n string.
@@ -736,21 +899,13 @@
         one that should coalesce.
         """
         self.assertOffsetSerialisation([], '',
-            self.smart_client, self.smart_server_request)
+            self.client_protocol, self.smart_server_request)
         self.assertOffsetSerialisation([(1,2)], '1,2',
-            self.smart_client, self.smart_server_request)
+            self.client_protocol, self.smart_server_request)
         self.assertOffsetSerialisation([(10,40), (0,5)], '10,40\n0,5',
-            self.smart_client, self.smart_server_request)
+            self.client_protocol, self.smart_server_request)
         self.assertOffsetSerialisation([(1,2), (3,4), (100, 200)],
-            '1,2\n3,4\n100,200', self.smart_client, self.smart_server_request)
-
-    def test_construct_version_one_protocol(self):
-        protocol = smart.SmartServerRequestProtocolOne(None, None)
-        self.assertEqual('', protocol.in_buffer)
-        self.assertFalse(protocol.has_dispatched)
-        # Once refactoring is complete, we don't need these assertions
-        self.assertFalse(hasattr(protocol, '_in'))
-        self.assertFalse(hasattr(protocol, '_out'))
+            '1,2\n3,4\n100,200', self.client_protocol, self.smart_server_request)
 
     def test_accept_bytes_to_protocol(self):
         out_stream = StringIO()
@@ -789,24 +944,126 @@
     def test_query_version(self):
         """query_version on a SmartClientProtocolOne should return a number.
         
-        XXX: SmartClientProtocolOne doesn't exist yet, so we pretend.
+        The protocol provides the query_version because the domain level clients
+        may all need to be able to probe for capabilities.
         """
         # What we really want to test here is that SmartClientProtocolOne calls
         # accept_bytes(tuple_based_encoding_of_hello) and reads and parses the
         # response of tuple-encoded (ok, 1).  Also, seperately we should test
-        # the error if the response a non-understood version.
-        input = StringIO('ok\x011\n' * 2)
-        output = StringIO()
-        medium = smart.SmartSimplePipesClientMedium(input, output)
-        # medium should really be a protocol eventually.
-        # protocol = smart.SmartClientProtocolOne(medium)
-        protocol = medium
-        self.assertEqual(1, protocol.query_version())
-        self.assertEqual(1, protocol.query_version())
-
+        # the error if the response is a non-understood version.
+        input = StringIO('ok\x011\n')
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        protocol = smart.SmartClientRequestProtocolOne(medium.get_request())
+        self.assertEqual(1, protocol.query_version())
+
+    def assertServerToClientEncoding(self, expected_bytes, expected_tuple,
+            input_tuples):
+        """Assert that each input_tuple serialises as expected_bytes, and the
+        bytes deserialise as expected_tuple.
+        """
+        # check the encoding of the server for all input_tuples matches
+        # expected bytes
+        for input_tuple in input_tuples:
+            server_output = StringIO()
+            server_protocol = smart.SmartServerRequestProtocolOne(server_output,
+                None)
+            server_protocol._send_response(input_tuple)
+            self.assertEqual(expected_bytes, server_output.getvalue())
+        # check the decoding of the client protocol from expected_bytes:
+        input = StringIO(expected_bytes)
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        protocol = smart.SmartClientRequestProtocolOne(medium.get_request())
+        protocol.call('foo')
+        self.assertEqual(expected_tuple, protocol.read_response_tuple())
+
+    def test_client_call_empty_response(self):
+        # protocol.call() can get back an empty tuple as a response. This occurs
+        # when the parsed line is an empty line, and results in a tuple with
+        # one element - an empty string.
+        self.assertServerToClientEncoding('\n', ('', ), [(), ('', )])
+
+    def test_client_call_three_element_response(self):
+        # protocol.call() can get back tuples of other lengths. A three element
+        # tuple should be unpacked as three strings.
+        self.assertServerToClientEncoding('a\x01b\x0134\n', ('a', 'b', '34'),
+            [('a', 'b', '34')])
+
+    def test_client_call_with_body_bytes_uploads(self):
+        # protocol.call_with_upload should length-prefix the bytes onto the 
+        # wire.
+        expected_bytes = "foo\n7\nabcdefgdone\n"
+        input = StringIO("\n")
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        protocol = smart.SmartClientRequestProtocolOne(medium.get_request())
+        protocol.call_with_body_bytes(('foo', ), "abcdefg")
+        self.assertEqual(expected_bytes, output.getvalue())
+
+    def test_client_call_with_body_readv_array(self):
+        # protocol.call_with_upload should encode the readv array and then
+        # length-prefix the bytes onto the wire.
+        expected_bytes = "foo\n7\n1,2\n5,6done\n"
+        input = StringIO("\n")
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        protocol = smart.SmartClientRequestProtocolOne(medium.get_request())
+        protocol.call_with_body_readv_array(('foo', ), [(1,2),(5,6)])
+        self.assertEqual(expected_bytes, output.getvalue())
+
+    def test_client_read_body_bytes_all(self):
+        # read_body_bytes should decode the body bytes from the wire into
+        # a response.
+        expected_bytes = "1234567"
+        server_bytes = "ok\n7\n1234567done\n"
+        input = StringIO(server_bytes)
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        protocol = smart.SmartClientRequestProtocolOne(medium.get_request())
+        protocol.call('foo')
+        protocol.read_response_tuple(True)
+        self.assertEqual(expected_bytes, protocol.read_body_bytes())
+
+    def test_client_read_body_bytes_incremental(self):
+        # test reading a few bytes at a time from the body
+        # XXX: possibly we should test dribbling the bytes into the stringio
+        # to make the state machine work harder: however, as we use the  
+        # LengthPrefixedBodyDecoder that is already well tested - we can skip
+        # that.
+        expected_bytes = "1234567"
+        server_bytes = "ok\n7\n1234567done\n"
+        input = StringIO(server_bytes)
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        protocol = smart.SmartClientRequestProtocolOne(medium.get_request())
+        protocol.call('foo')
+        protocol.read_response_tuple(True)
+        self.assertEqual(expected_bytes[0:2], protocol.read_body_bytes(2))
+        self.assertEqual(expected_bytes[2:4], protocol.read_body_bytes(2))
+        self.assertEqual(expected_bytes[4:6], protocol.read_body_bytes(2))
+        self.assertEqual(expected_bytes[6], protocol.read_body_bytes())
+
+    def test_client_cancel_read_body_does_not_eat_body_bytes(self):
+        # cancelling the expected body needs to finish the request, but not
+        # read any more bytes.
+        expected_bytes = "1234567"
+        server_bytes = "ok\n7\n1234567done\n"
+        input = StringIO(server_bytes)
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        protocol = smart.SmartClientRequestProtocolOne(medium.get_request())
+        protocol.call('foo')
+        protocol.read_response_tuple(True)
+        protocol.cancel_read_body()
+        self.assertEqual(3, input.tell())
+        self.assertRaises(errors.ReadingCompleted, protocol.read_body_bytes)
 
 class LengthPrefixedBodyDecoder(tests.TestCase):
 
+    # XXX: TODO: make accept_reading_trailer invoke translate_response or 
+    # something similar to the ProtocolBase method.
+
     def test_construct(self):
         decoder = smart.LengthPrefixedBodyDecoder()
         self.assertFalse(decoder.finished_reading)
@@ -874,17 +1131,6 @@
         remote_transport = RemoteTransport('fake_url', medium)
         self.assertEqual("c", remote_transport.readv("data-file", [(0,1)]))
 
-        
-##        client = http_transport.get_smart_client()
-##        # XXX: this should be a domain method, not a call to a structured data
-##        # method, e.g. "client.read_bundle(...)"
-##        answer = client._call_with_upload(
-##            'readv', ('data-file',), '0,1')
-##        answer_body = client._recv_bulk()
-##
-##        self.assertEqual(('readv',), answer)
-##        self.assertEqual('c', answer_body)
-        
     def test_http_server_with_smarts(self):
         http_server = HTTPServerWithSmarts()
         http_server.setUp()

=== modified file 'bzrlib/tests/test_transport_implementations.py'
--- bzrlib/tests/test_transport_implementations.py	2006-09-21 09:15:20 +0000
+++ bzrlib/tests/test_transport_implementations.py	2006-09-22 04:39:58 +0000
@@ -1311,22 +1311,6 @@
         self.assertEqual(d[2], (0, '0'))
         self.assertEqual(d[3], (3, '34'))
 
-    def test_get_smart_client(self):
-        """All transports must either give a smart client, or know they can't.
-
-        For some transports such as http this might depend on probing to see 
-        what's actually present on the other end.  (But we can adjust for that 
-        in the future.)
-        """
-        transport = self.get_transport()
-        try:
-            client = transport.get_smart_client()
-            # XXX: should be a more general class
-            self.assertIsInstance(client, smart.SmartStreamClient)
-        except NoSmartServer:
-            # as long as we got it we're fine
-            pass
-
     def test_get_smart_medium(self):
         """All transports must either give a smart medium, or know they can't.
         """

=== modified file 'bzrlib/transport/smart.py'
--- bzrlib/transport/smart.py	2006-09-21 09:15:20 +0000
+++ bzrlib/transport/smart.py	2006-09-22 04:39:58 +0000
@@ -249,19 +249,11 @@
 class SmartProtocolBase(object):
     """Methods common to client and server"""
 
-    def _send_bulk_data(self, body, a_file=None):
-        """Send chunked body data"""
-        assert isinstance(body, str)
-        bytes = self._encode_bulk_data(body)
-        self._write_and_flush(bytes, a_file)
-
-    def _encode_bulk_data(self, body):
-        """Encode body as a bulk data chunk."""
-        return ''.join(('%d\n' % len(body), body, 'done\n'))
-
     # TODO: this only actually accomodates a single block; possibly should support
     # multiple chunks?
     def _recv_bulk(self):
+        # This is OBSOLETE except for the double handline in the server: 
+        # the read_bulk + reencode noise.
         chunk_len = self._in.readline()
         try:
             chunk_len = int(chunk_len)
@@ -273,9 +265,6 @@
         self._recv_trailer()
         return bulk
 
-    def _recv_tuple(self):
-        return _recv_tuple(self._in)
-
     def _recv_trailer(self):
         resp = self._recv_tuple()
         if resp == ('done', ):
@@ -283,6 +272,10 @@
         else:
             self._translate_error(resp)
 
+    def _encode_bulk_data(self, body):
+        """Encode body as a bulk data chunk."""
+        return ''.join(('%d\n' % len(body), body, 'done\n'))
+
     def _serialise_offsets(self, offsets):
         """Serialise a readv offset list."""
         txt = []
@@ -290,6 +283,12 @@
             txt.append('%d,%d' % (start, length))
         return '\n'.join(txt)
 
+    def _send_bulk_data(self, body, a_file=None):
+        """Send chunked body data"""
+        assert isinstance(body, str)
+        bytes = self._encode_bulk_data(body)
+        self._write_and_flush(bytes, a_file)
+
     def _write_and_flush(self, bytes, a_file=None):
         """Write bytes to self._out and flush it."""
         # XXX: this will be inefficient.  Just ask Robert.
@@ -297,7 +296,7 @@
             a_file = self._out
         a_file.write(bytes)
         a_file.flush()
-
+        
 
 class SmartServerRequestProtocolOne(SmartProtocolBase):
     """Server-side encoding and decoding logic for smart version 1."""
@@ -380,7 +379,7 @@
     def sync_with_request(self, request):
         self.finished_reading = request.finished_reading
         
-        
+
 class LengthPrefixedBodyDecoder(object):
     """Decodes the length-prefixed bulk data."""
     
@@ -450,10 +449,7 @@
         result = self._in_buffer
         self._in_buffer = ''
         return result
-            
-        
 
-        
 
 class SmartServerStreamMedium(SmartProtocolBase):
     """Handles smart commands coming over a stream.
@@ -932,14 +928,14 @@
         self._scheme, self._username, self._password, self._host, self._port, self._path = \
                 transport.split_url(url)
         if clone_from is None:
-            self._client = medium
+            self._medium = medium
         else:
             # credentials may be stripped from the base in some circumstances
             # as yet to be clearly defined or documented, so copy them.
             self._username = clone_from._username
             # reuse same connection
-            self._client = clone_from._client
-        assert self._client is not None
+            self._medium = clone_from._medium
+        assert self._medium is not None
 
     def abspath(self, relpath):
         """Return the full url to the given relative path.
@@ -964,10 +960,10 @@
         return False
                                                    
     def get_smart_client(self):
-        return self._client
+        return self._medium
 
     def get_smart_medium(self):
-        return self._client
+        return self._medium
                                                    
     def _unparse_url(self, path):
         """Return URL for a path.
@@ -990,12 +986,28 @@
         """Returns the Unicode version of the absolute path for relpath."""
         return self._combine_paths(self._path, relpath)
 
+    def _call(self, method, *args):
+        resp = self._call2(method, *args)
+        self._translate_error(resp)
+
+    def _call2(self, method, *args):
+        """Call a method on the remote server."""
+        protocol = SmartClientRequestProtocolOne(self._medium.get_request())
+        protocol.call(method, *args)
+        return protocol.read_response_tuple()
+
+    def _call_with_body_bytes(self, method, args, body):
+        """Call a method on the remote server with body bytes."""
+        protocol = SmartClientRequestProtocolOne(self._medium.get_request())
+        protocol.call_with_body_bytes((method, ) + args, body)
+        return protocol.read_response_tuple()
+
     def has(self, relpath):
         """Indicate whether a remote file of the given name exists or not.
 
         :see: Transport.has()
         """
-        resp = self._client._call('has', self._remote_path(relpath))
+        resp = self._call2('has', self._remote_path(relpath))
         if resp == ('yes', ):
             return True
         elif resp == ('no', ):
@@ -1008,11 +1020,17 @@
         
         :see: Transport.get_bytes()/get_file()
         """
+        return StringIO(self.get_bytes(relpath))
+
+    def get_bytes(self, relpath):
         remote = self._remote_path(relpath)
-        resp = self._client._call('get', remote)
+        protocol = SmartClientRequestProtocolOne(self._medium.get_request())
+        protocol.call('get', remote)
+        resp = protocol.read_response_tuple(True)
         if resp != ('ok', ):
+            protocol.cancel_read_body()
             self._translate_error(resp, relpath)
-        return StringIO(self._client._recv_bulk())
+        return protocol.read_body_bytes()
 
     def _serialise_optional_mode(self, mode):
         if mode is None:
@@ -1021,17 +1039,15 @@
             return '%d' % mode
 
     def mkdir(self, relpath, mode=None):
-        resp = self._client._call('mkdir', 
-                                  self._remote_path(relpath), 
-                                  self._serialise_optional_mode(mode))
+        resp = self._call2('mkdir', self._remote_path(relpath),
+            self._serialise_optional_mode(mode))
         self._translate_error(resp)
 
     def put_bytes(self, relpath, upload_contents, mode=None):
         # FIXME: upload_file is probably not safe for non-ascii characters -
         # should probably just pass all parameters as length-delimited
         # strings?
-        resp = self._client._call_with_upload(
-            'put',
+        resp = self._call_with_body_bytes('put',
             (self._remote_path(relpath), self._serialise_optional_mode(mode)),
             upload_contents)
         self._translate_error(resp)
@@ -1045,7 +1061,7 @@
         if create_parent_dir:
             create_parent_str = 'T'
 
-        resp = self._client._call_with_upload(
+        resp = self._call_with_body_bytes(
             'put_non_atomic',
             (self._remote_path(relpath), self._serialise_optional_mode(mode),
              create_parent_str, self._serialise_optional_mode(dir_mode)),
@@ -1074,7 +1090,7 @@
         return self.append_bytes(relpath, from_file.read(), mode)
         
     def append_bytes(self, relpath, bytes, mode=None):
-        resp = self._client._call_with_upload(
+        resp = self._call_with_body_bytes(
             'append',
             (self._remote_path(relpath), self._serialise_optional_mode(mode)),
             bytes)
@@ -1083,7 +1099,7 @@
         self._translate_error(resp)
 
     def delete(self, relpath):
-        resp = self._client._call('delete', self._remote_path(relpath))
+        resp = self._call2('delete', self._remote_path(relpath))
         self._translate_error(resp)
 
     def readv(self, relpath, offsets):
@@ -1100,18 +1116,20 @@
                                limit=self._max_readv_combine,
                                fudge_factor=self._bytes_to_read_before_seek))
 
-
-        resp = self._client._call_with_upload(
-            'readv',
-            (self._remote_path(relpath),),
-            self._client._serialise_offsets((c.start, c.length) for c in coalesced))
+        protocol = SmartClientRequestProtocolOne(self._medium.get_request())
+        protocol.call_with_body_readv_array(
+            ('readv', self._remote_path(relpath)),
+            [(c.start, c.length) for c in coalesced])
+        resp = protocol.read_response_tuple(True)
 
         if resp[0] != 'readv':
             # This should raise an exception
+            protocol.cancel_read_body()
             self._translate_error(resp)
             return
 
-        data = self._client._recv_bulk()
+        # FIXME: this should know how many bytes are needed, for clarity.
+        data = protocol.read_body_bytes()
         # Cache the results, but only until they have been fulfilled
         data_map = {}
         for c_offset in coalesced:
@@ -1130,22 +1148,18 @@
                 cur_offset_and_size = offset_stack.next()
 
     def rename(self, rel_from, rel_to):
-        self._call('rename', 
+        self._call('rename',
                    self._remote_path(rel_from),
                    self._remote_path(rel_to))
 
     def move(self, rel_from, rel_to):
-        self._call('move', 
+        self._call('move',
                    self._remote_path(rel_from),
                    self._remote_path(rel_to))
 
     def rmdir(self, relpath):
         resp = self._call('rmdir', self._remote_path(relpath))
 
-    def _call(self, method, *args):
-        resp = self._client._call(method, *args)
-        self._translate_error(resp)
-
     def _translate_error(self, resp, orig_path=None):
         """Raise an exception from a response"""
         if resp is None:
@@ -1188,20 +1202,14 @@
         else:
             raise errors.SmartProtocolError('unexpected smart server error: %r' % (resp,))
 
-    def _send_tuple(self, args):
-        self._client.accept_bytes(_encode_tuple(args))
-
-    def _recv_tuple(self):
-        return self._client._recv_tuple()
-
     def disconnect(self):
-        self._client.disconnect()
+        self._medium.disconnect()
 
     def delete_tree(self, relpath):
         raise errors.TransportNotPossible('readonly transport')
 
     def stat(self, relpath):
-        resp = self._client._call('stat', self._remote_path(relpath))
+        resp = self._call2('stat', self._remote_path(relpath))
         if resp[0] == 'stat':
             return SmartStat(int(resp[1]), int(resp[2], 8))
         else:
@@ -1224,116 +1232,360 @@
         return True
 
     def list_dir(self, relpath):
-        resp = self._client._call('list_dir',
-                                  self._remote_path(relpath))
+        resp = self._call2('list_dir', self._remote_path(relpath))
         if resp[0] == 'names':
             return [name.encode('ascii') for name in resp[1:]]
         else:
             self._translate_error(resp)
 
     def iter_files_recursive(self):
-        resp = self._client._call('iter_files_recursive',
-                                  self._remote_path(''))
+        resp = self._call2('iter_files_recursive', self._remote_path(''))
         if resp[0] == 'names':
             return resp[1:]
         else:
             self._translate_error(resp)
 
 
-class SmartStreamClient(SmartProtocolBase):
-    """Connection to smart server over two streams"""
-
-    def _send_bulk_data(self, body):
-        self._ensure_connection()
-        SmartProtocolBase._send_bulk_data(self, body)
-        
-    def disconnect(self):
-        """If this medium maintains a persistent connection, close it.
-        
-        The default implementation does nothing.
-        """
-        
-    def _call(self, *args):
-        bytes = _encode_tuple(args)
-        # should be self.medium.accept_bytes(bytes) XXX
-        self.accept_bytes(bytes)
-        return self._recv_tuple()
-
-    def _call_with_upload(self, method, args, body):
-        """Call an rpc, supplying bulk upload data.
-
-        :param method: method name to call
-        :param args: parameter args tuple
-        :param body: upload body as a byte string
-        """
-        bytes = _encode_tuple((method,) + args)
-        bytes += self._encode_bulk_data(body)
-        # should be self.medium.accept_bytes XXX
-        self.accept_bytes(bytes)
-        return self._recv_tuple()
+class SmartClientMediumRequest(object):
+    """A request on a SmartClientMedium.
+
+    Each request allows bytes to be provided to it via accept_bytes, and then
+    the response bytes to be read via read_bytes.
+
+    For instance:
+    request.accept_bytes('123')
+    request.finished_writing()
+    result = request.read_bytes(3)
+    request.finished_reading()
+
+    It is up to the individual SmartClientMedium whether multiple concurrent
+    requests can exist. See SmartClientMedium.get_request to obtain instances 
+    of SmartClientMediumRequest, and the concrete Medium you are using for 
+    details on concurrency and pipelining.
+    """
+
+    def __init__(self, medium):
+        """Construct a SmartClientMediumRequest for the medium medium."""
+        self._medium = medium
+        # we track state by constants - we may want to use the same
+        # pattern as BodyReader if it gets more complex.
+        # valid states are: "writing", "reading", "done"
+        self._state = "writing"
+
+    def accept_bytes(self, bytes):
+        """Accept bytes for inclusion in this request.
+
+        This method may not be be called after finished_writing() has been
+        called.  It depends upon the Medium whether or not the bytes will be
+        immediately transmitted. Message based Mediums will tend to buffer the
+        bytes until finished_writing() is called.
+
+        :param bytes: A bytestring.
+        """
+        if self._state != "writing":
+            raise errors.WritingCompleted(self)
+        self._accept_bytes(bytes)
+
+    def _accept_bytes(self, bytes):
+        """Helper for accept_bytes.
+
+        Accept_bytes checks the state of the request to determing if bytes
+        should be accepted. After that it hands off to _accept_bytes to do the
+        actual acceptance.
+        """
+        raise NotImplementedError(self._accept_bytes)
+
+    def finished_reading(self):
+        """Inform the request that all desired data has been read.
+
+        This will remove the request from the pipeline for its medium (if the
+        medium supports pipelining) and any further calls to methods on the
+        request will raise ReadingCompleted.
+        """
+        if self._state == "writing":
+            raise errors.WritingNotComplete(self)
+        if self._state != "reading":
+            raise errors.ReadingCompleted(self)
+        self._state = "done"
+        self._finished_reading()
+
+    def _finished_reading(self):
+        """Helper for finished_reading.
+
+        finished_reading checks the state of the request to determine if 
+        finished_reading is allowed, and if it is hands off to _finished_reading
+        to perform the action.
+        """
+        raise NotImplementedError(self._finished_reading)
+
+    def finished_writing(self):
+        """Finish the writing phase of this request.
+
+        This will flush all pending data for this request along the medium.
+        After calling finished_writing, you may not call accept_bytes anymore.
+        """
+        if self._state != "writing":
+            raise errors.WritingCompleted(self)
+        self._state = "reading"
+        self._finished_writing()
+
+    def _finished_writing(self):
+        """Helper for finished_writing.
+
+        finished_writing checks the state of the request to determine if 
+        finished_writing is allowed, and if it is hands off to _finished_writing
+        to perform the action.
+        """
+        raise NotImplementedError(self._finished_writing)
+
+    def read_bytes(self, count):
+        """Read bytes from this requests response.
+
+        This method will block and wait for count bytes to be read. It may not
+        be invoked until finished_writing() has been called - this is to ensure
+        a message-based approach to requests, for compatability with message
+        based mediums like HTTP.
+        """
+        if self._state == "writing":
+            raise errors.WritingNotComplete(self)
+        if self._state != "reading":
+            raise errors.ReadingCompleted(self)
+        return self._read_bytes(count)
+
+    def _read_bytes(self, count):
+        """Helper for read_bytes.
+
+        read_bytes checks the state of the request to determing if bytes
+        should be read. After that it hands off to _read_bytes to do the
+        actual read.
+        """
+        raise NotImplementedError(self._read_bytes)
+
+
+class SmartClientStreamMediumRequest(SmartClientMediumRequest):
+    """A SmartClientMediumRequest that works with an SmartClientStreamMedium."""
+
+    def __init__(self, medium):
+        SmartClientMediumRequest.__init__(self, medium)
+        # check that we are safe concurrency wise. If some streams start
+        # allowing concurrent requests - i.e. via multiplexing - then this
+        # assert should be moved to SmartClientStreamMedium.get_request,
+        # and the setting/unsetting of _current_request likewise moved into
+        # that class : but its unneeded overhead for now. RBC 20060922
+        if self._medium._current_request is not None:
+            raise errors.TooManyConcurrentRequests(self._medium)
+        self._medium._current_request = self
+
+    def _accept_bytes(self, bytes):
+        """See SmartClientMediumRequest._accept_bytes.
+        
+        This forwards to self._medium._accept_bytes because we are operating
+        on the mediums stream.
+        """
+        self._medium._accept_bytes(bytes)
+
+    def _finished_reading(self):
+        """See SmartClientMediumRequest._finished_reading.
+
+        This clears the _current_request on self._medium to allow a new 
+        request to be created.
+        """
+        assert self._medium._current_request is self
+        self._medium._current_request = None
+        
+    def _finished_writing(self):
+        """See SmartClientMediumRequest._finished_writing.
+
+        This invokes self._medium._flush to ensure all bytes are transmitted.
+        """
+        self._medium._flush()
+
+    def _read_bytes(self, count):
+        """See SmartClientMediumRequest._read_bytes.
+        
+        This forwards to self._medium._read_bytes because we are operating
+        on the mediums stream.
+        """
+        return self._medium._read_bytes(count)
+
+
+class SmartClientRequestProtocolOne(SmartProtocolBase):
+    """The client-side protocol for smart version 1."""
+
+    def __init__(self, request):
+        """Construct a SmartClientRequestProtocolOne.
+
+        :param request: A SmartClientMediumRequest to serialise onto and
+            deserialise from.
+        """
+        self._request = request
+        self._body_buffer = None
+
+    def call(self, *args):
+        bytes = _encode_tuple(args)
+        self._request.accept_bytes(bytes)
+        self._request.finished_writing()
+
+    def call_with_body_bytes(self, args, body):
+        """Make a remote call of args with body bytes 'body'.
+
+        After calling this, call read_response_tuple to find the result out.
+        """
+        bytes = _encode_tuple(args)
+        self._request.accept_bytes(bytes)
+        bytes = self._encode_bulk_data(body)
+        self._request.accept_bytes(bytes)
+        self._request.finished_writing()
+
+    def call_with_body_readv_array(self, args, body):
+        """Make a remote call with a readv array.
+
+        The body is encoded with one line per readv offset pair. The numbers in
+        each pair are separated by a comma, and no trailing \n is emitted.
+        """
+        bytes = _encode_tuple(args)
+        self._request.accept_bytes(bytes)
+        readv_bytes = self._serialise_offsets(body)
+        bytes = self._encode_bulk_data(readv_bytes)
+        self._request.accept_bytes(bytes)
+        self._request.finished_writing()
+
+    def cancel_read_body(self):
+        """After expecting a body, a response code may indicate one otherwise.
+
+        This method lets the domain client inform the protocol that no body
+        will be transmitted. This is a terminal method: after calling it the
+        protocol is not able to be used further.
+        """
+        self._request.finished_reading()
+
+    def read_response_tuple(self, expect_body=False):
+        """Read a response tuple from the wire.
+
+        This should only be called once.
+        """
+        result = self._recv_tuple()
+        if not expect_body:
+            self._request.finished_reading()
+        return result
+
+    def read_body_bytes(self, count=-1):
+        """Read bytes from the body, decoding into a byte stream.
+        
+        We read all bytes at once to ensure we've checked the trailer for 
+        errors, and then feed the buffer back as read_body_bytes is called.
+        """
+        if self._body_buffer is not None:
+            return self._body_buffer.read(count)
+        _body_decoder = LengthPrefixedBodyDecoder()
+        # grab a byte from the wire: we do this so that we dont use too much
+        # from the wire; we should have the LengthPrefixedBodyDecoder tell
+        # us how much is needed once the header is written.
+        # i.e. self._body_decoder.next_read_size() would be a hint. 
+        while not _body_decoder.finished_reading:
+            byte = self._request.read_bytes(1)
+            _body_decoder.accept_bytes(byte)
+        self._request.finished_reading()
+        self._body_buffer = StringIO(_body_decoder.read_pending_data())
+        # XXX: TODO check the trailer result.
+        return self._body_buffer.read(count)
+
+    def _recv_tuple(self):
+        """Recieve a tuple from the medium request."""
+        line = ''
+        while not line or line[-1] != '\n':
+            # yes, this is inefficient - but tuples are short.
+            new_char = self._request.read_bytes(1)
+            line += new_char
+            assert new_char != '', "end of file reading from server."
+        return _decode_tuple(line)
 
     def query_version(self):
         """Return protocol version number of the server."""
-        resp = self._call('hello')
+        self.call('hello')
+        resp = self.read_response_tuple()
         if resp == ('ok', '1'):
             return 1
         else:
             raise errors.SmartProtocolError("bad response %r" % (resp,))
 
 
-class SmartClientMedium(SmartStreamClient):
-    """Smart client is a medium for sending smart protocol requests over.
-
-    XXX: we want explicit finalisation
-    """
-
-
-class SmartStreamMediumClient(SmartClientMedium):
-    """The SmartStreamMediumClient knows how to close the stream when it is
-    finished with it.
-    """
+class SmartClientMedium(object):
+    """Smart client is a medium for sending smart protocol requests over."""
+
+    def disconnect(self):
+        """If this medium maintains a persistent connection, close it.
+        
+        The default implementation does nothing.
+        """
+        
+
+class SmartClientStreamMedium(SmartClientMedium):
+    """Stream based medium common class.
+
+    SmartClientStreamMediums operate on a stream. All subclasses use a common
+    SmartClientStreamMediumRequest for their requests, and should implement
+    _accept_bytes and _read_bytes to allow the request objects to send and
+    receive bytes.
+    """
+
+    def __init__(self):
+        self._current_request = None
+
+    def accept_bytes(self, bytes):
+        self._accept_bytes(bytes)
 
     def __del__(self):
+        """The SmartClientStreamMedium knows how to close the stream when it is
+        finished with it.
+        """
         self.disconnect()
 
-
-class SmartSimplePipesClientMedium(SmartClientMedium):
+    def _flush(self):
+        """Flush the output stream.
+        
+        This method is used by the SmartClientStreamMediumRequest to ensure that
+        all data for a request is sent, to avoid long timeouts or deadlocks.
+        """
+        raise NotImplementedError(self._flush)
+
+    def get_request(self):
+        """See SmartClientMedium.get_request().
+
+        SmartClientStreamMedium always returns a SmartClientStreamMediumRequest
+        for get_request.
+        """
+        return SmartClientStreamMediumRequest(self)
+
+    def read_bytes(self, count):
+        return self._read_bytes(count)
+
+
+class SmartSimplePipesClientMedium(SmartClientStreamMedium):
     """A client medium using simple pipes.
     
     This client does not manage the pipes: it assumes they will always be open.
     """
 
     def __init__(self, readable_pipe, writeable_pipe):
+        SmartClientStreamMedium.__init__(self)
         self._readable_pipe = readable_pipe
         self._writeable_pipe = writeable_pipe
 
-    def accept_bytes(self, bytes):
-        """See SmartClientMedium.accept_bytes."""
+    def _accept_bytes(self, bytes):
+        """See SmartClientStreamMedium.accept_bytes."""
         self._writeable_pipe.write(bytes)
 
-    def read_bytes(self, count):
-        """See SmartClientMedium.read_bytes."""
+    def _flush(self):
+        """See SmartClientStreamMedium._flush()."""
+        self._writeable_pipe.flush()
+
+    def _read_bytes(self, count):
+        """See SmartClientStreamMedium._read_bytes."""
         return self._readable_pipe.read(count)
 
-    def _recv_bulk(self):
-        """transitional api from 'client' to 'medium'."""
-        self._in = self._readable_pipe
-        try:
-            return SmartClientMedium._recv_bulk(self)
-        finally:
-            self._in = None
-
-    def _recv_tuple(self):
-        """transitional api from 'client' to 'medium'."""
-        return _recv_tuple(self._readable_pipe)
-
-    def _write_and_flush(self, bytes, file=None):
-        """Thunk from the 'client' api to the 'Medium' api."""
-        assert file is None
-        self.accept_bytes(bytes)
-
-
-class SmartSSHClientMedium(SmartStreamMediumClient):
+
+class SmartSSHClientMedium(SmartClientStreamMedium):
     """A client medium using SSH."""
     
     def __init__(self, host, port=None, username=None, password=None,
@@ -1343,6 +1595,7 @@
         :param vendor: An optional override for the ssh vendor to use. See
             bzrlib.transport.ssh for details on ssh vendors.
         """
+        SmartClientStreamMedium.__init__(self)
         self._connected = False
         self._host = host
         self._password = password
@@ -1353,11 +1606,10 @@
         self._vendor = vendor
         self._write_to = None
 
-    def accept_bytes(self, bytes):
-        """See SmartClientMedium.accept_bytes."""
+    def _accept_bytes(self, bytes):
+        """See SmartClientStreamMedium.accept_bytes."""
         self._ensure_connection()
         self._write_to.write(bytes)
-        self._write_to.flush()
 
     def disconnect(self):
         """See SmartClientMedium.disconnect()."""
@@ -1385,39 +1637,29 @@
             self._ssh_connection.get_filelike_channels()
         self._connected = True
 
-    def read_bytes(self, count):
-        """See SmartClientMedium.read_bytes."""
-        raise errors.MediumNotConnected(self)
-
-    def _recv_bulk(self):
-        """transitional api from 'client' to 'medium'."""
-        self._in = self._read_from
-        try:
-            return SmartStreamMediumClient._recv_bulk(self)
-        finally:
-            self._in = None
-
-    def _recv_tuple(self):
-        """transitional api from 'client' to 'medium'."""
-        return _recv_tuple(self._read_from)
-
-    def _write_and_flush(self, bytes, file=None):
-        """Thunk from the 'client' api to the 'Medium' api."""
-        assert file is None
-        self.accept_bytes(bytes)
-
-
-class SmartTCPClientMedium(SmartStreamMediumClient):
+    def _flush(self):
+        """See SmartClientStreamMedium._flush()."""
+        self._write_to.flush()
+
+    def _read_bytes(self, count):
+        """See SmartClientStreamMedium.read_bytes."""
+        if not self._connected:
+            raise errors.MediumNotConnected(self)
+        return self._read_from.read(count)
+
+
+class SmartTCPClientMedium(SmartClientStreamMedium):
     """A client medium using TCP."""
     
     def __init__(self, host, port):
         """Creates a client that will connect on the first use."""
+        SmartClientStreamMedium.__init__(self)
         self._connected = False
         self._host = host
         self._port = port
         self._socket = None
 
-    def accept_bytes(self, bytes):
+    def _accept_bytes(self, bytes):
         """See SmartClientMedium.accept_bytes."""
         self._ensure_connection()
         self._socket.sendall(bytes)
@@ -1442,26 +1684,18 @@
                     (self._host, self._port, os.strerror(result)))
         self._connected = True
 
-    def read_bytes(self, count):
+    def _flush(self):
+        """See SmartClientStreamMedium._flush().
+        
+        For TCP we do no flushing. We may want to turn off TCP_NODELAY and 
+        add a means to do a flush, but that can be done in the future.
+        """
+
+    def _read_bytes(self, count):
         """See SmartClientMedium.read_bytes."""
-        raise errors.MediumNotConnected(self)
-
-    def _recv_bulk(self):
-        """transitional api from 'client' to 'medium'."""
-        self._in = self._socket.makefile('r', 0)
-        try:
-            return SmartStreamMediumClient._recv_bulk(self)
-        finally:
-            self._in = None
-
-    def _recv_tuple(self):
-        """transitional api from 'client' to 'medium'."""
-        return _recv_tuple(self._socket.makefile('r', 0))
-
-    def _write_and_flush(self, bytes, file=None):
-        """Thunk from the 'client' api to the 'Medium' api."""
-        assert file is None
-        self.accept_bytes(bytes)
+        if not self._connected:
+            raise errors.MediumNotConnected(self)
+        return self._socket.recv(count)
 
 
 class SmartTCPTransport(SmartTransport):

